/*
Copyright (C) September 1999, Ka-Ping Yee

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "dynim.h"

/* DynImCanvas implementation ========================================== */

/* -------------------------------------------------------- construction */
DynImCanvas* dynim_canvas_new(DynImPalette* palette, int width, int height)
    {
    DynImCanvas* canvas = alloc(DynImCanvas);
    canvas->palette = palette;
    canvas->width = width;
    canvas->height = height;
    canvas->data = allocarray(byte, width * height);
    return canvas;
    }

void dynim_canvas_delete(DynImCanvas* canvas)
    {
    toast(canvas->data);
    toast(canvas);
    }

/* ------------------------------------------------------------- palette */
void dynim_canvas_setpalette(DynImCanvas* canvas, DynImPalette* palette)
    {
    canvas->palette = palette;
    }

/* ------------------------------------------------------------- drawing */
void dynim_canvas_clear(DynImCanvas* canvas, byte colour)
    {
    memset(canvas->data, colour, canvas->width * canvas->height);
    }

void dynim_canvas_drawrect(DynImCanvas* canvas, byte colour,
                           int a, int b, int x, int y)
    {
    int w = canvas->width, h = canvas->height;
    int xmin = (a < x) ? a : x;
    int ymin = (b < y) ? b : y;
    int xmax = (a > x) ? a : x;
    int ymax = (b > y) ? b : y;
    int xminok = (xmin >= 0 && xmin < w);
    int xmaxok = (xmax >= 0 && xmax < w);
    int yminok = (ymin >= 0 && ymin < h);
    int ymaxok = (ymax >= 0 && ymax < h);
    byte* pixel;

    xmin = xmin < 0 ? 0 : xmin >= w ? w-1 : xmin;
    xmax = xmax < 0 ? 0 : xmax >= w ? w-1 : xmax;
    ymin = ymin < 0 ? 0 : ymin >= h ? h-1 : ymin;
    ymax = ymax < 0 ? 0 : ymax >= h ? h-1 : ymax;

    if (xminok)
        {
        pixel = dynim_canvas_pixelptr(canvas, xmin, ymin);
        for (y = ymin; y <= ymax; y++, pixel+=w)
            *pixel = colour;
        }

    if (xmaxok)
        {
        pixel = dynim_canvas_pixelptr(canvas, xmax, ymin);
        for (y = ymin; y <= ymax; y++, pixel+=w)
            *pixel = colour;
        }

    if (yminok)
        {
        pixel = dynim_canvas_pixelptr(canvas, xmin, ymin);
        for (x = xmin; x <= xmax; x++, pixel++)
            *pixel = colour;
        }

    if (ymaxok)
        {
        pixel = dynim_canvas_pixelptr(canvas, xmin, ymax);
        for (x = xmin; x <= xmax; x++, pixel++)
            *pixel = colour;
        }
    }

void dynim_canvas_fillrect(DynImCanvas* canvas, byte colour,
                           int a, int b, int x, int y)
    {
    int w = canvas->width, h = canvas->height;
    int xmin = (a < x) ? a : x;
    int ymin = (b < y) ? b : y;
    int xmax = (a > x) ? a : x;
    int ymax = (b > y) ? b : y;
    byte* pixel;
    byte* rowstart;

    xmin = xmin < 0 ? 0 : xmin >= w ? w-1 : xmin;
    xmax = xmax < 0 ? 0 : xmax >= w ? w-1 : xmax;
    ymin = ymin < 0 ? 0 : ymin >= h ? h-1 : ymin;
    ymax = ymax < 0 ? 0 : ymax >= h ? h-1 : ymax;

    rowstart = dynim_canvas_pixelptr(canvas, xmin, ymin);
    for (y = ymin; y <= ymax; y++)
        {
        pixel = rowstart;
        for (x = xmin; x <= xmax; x++, pixel++)
            *pixel = colour;
        rowstart += w;
        }
    }

int dynim_canvas_drawtext(DynImCanvas* canvas, DynImFont* font,
                          char* text, int x, int y, int shift)
    {
    DynImMetric* metric;
    int fx, fy;
    int cxmin, cxmax, cymin, cymax, cx, cy;
    byte* fpixel;
    byte* pixel;

    for (; *text; text++)
        {
        metric = dynim_font_getmetric(font, *text);

        fy = metric->top;
        cymin = y + metric->top - metric->baseline;
        if (cymin < 0)
            {
            fy += (-cymin);
            cymin = 0;
            }
        cymax = y + metric->bottom - metric->baseline;
        if (cymax > canvas->height)
            {
            cymax = canvas->height;
            }

        fx = metric->left;
        cxmin = x + metric->left - metric->origin;
        if (cxmin < 0)
            {
            fx += (-cxmin);
            cxmin = 0;
            }
        cxmax = x + metric->right - metric->origin;
        if (cxmax > canvas->width)
            {
            cxmax = canvas->width;
            }

        for (cy = cymin; cy < cymax; cy++, fy++)
            {
            fpixel = dynim_canvas_pixelptr(font->canvas, fx, fy);
            pixel = dynim_canvas_pixelptr(canvas, cxmin, cy);
            for (cx = cxmin; cx < cxmax; cx++)
                {
                if (*fpixel) *pixel = *fpixel + shift;
                fpixel++;
                pixel++;
                }
            }

        x += metric->offset;
        }

    return x;
    }

/* ------------------------------------------------------------- copying */
void read_line(int handle, char* line, int max)
    {
    char* ch;
    ch = line;
    while (ch - line < max - 1)
        {
        read(handle, ch, 1);
        if (*ch == '\n') break;
        ch++;
        }
    *ch = 0;
    }

DynImCanvas* dynim_canvas_loadppm(int handle)
    {
    DynImCanvas* canvas;
    DynImPalette* palette = dynim_palette_new(256);
    DynImColour colour;
    unsigned char* pixel;
    char line[80];
    int width, height, depth, x, c;

    read_line(handle, line, 80);
    if (line[0] != 'P' || line[1] != '6') return NULL;

    do read_line(handle, line, 80);
    while (line[0] == '#');
    sscanf(line, "%d %d", &width, &height);

    do read_line(handle, line, 80);
    while (line[0] == '#');
    sscanf(line, "%d", &depth);

    canvas = dynim_canvas_new(palette, width, height);
    pixel = canvas->data;
    palette->size = 0;

    for (x = 0; x < width * height; x++)
        {
        read(handle, &colour, 3);
        for (c = 0; c < palette->size; c++)
            {
            if (palette->colours[c].r == colour.r &&
                palette->colours[c].g == colour.g &&
                palette->colours[c].b == colour.b) break;
            }
        if (c >= palette->size)
            {
            palette->colours[c] = colour;
            palette->size++;
            }

        *pixel = c;
        pixel++;
        }

    return canvas;
    }

DynImCanvas* dynim_canvas_clone(DynImCanvas* source)
    {
    DynImCanvas* canvas =
        dynim_canvas_new(source->palette, source->width, source->height);
    memcpy(canvas->data, source->data, source->width * source->height);
    return canvas;
    }

DynImCanvas* dynim_canvas_copy(DynImCanvas* source, byte fill,
                               int l, int r, int t, int b)
    {
    int width, height;
    int sxmin, sxmax, symin, symax, sx, sy;
    int cx, cy;
    byte* spixel;
    byte* pixel;
    DynImCanvas* canvas;

    width = (r - l > 0) ? r - l : 0;
    height = (b - t > 0) ? b - t : 0;

    canvas = dynim_canvas_new(source->palette, width, height);
    dynim_canvas_clear(canvas, fill);

    cy = 0;
    symin = t;
    if (symin < 0)
        {
        cy += (-symin);
        symin = 0;
        }
    symax = b;
    if (symax > source->height)
        {
        symax = source->height;
        }

    cx = 0;
    sxmin = l;
    if (sxmin < 0)
        {
        cx += (-sxmin);
        sxmin = 0;
        }
    sxmax = r;
    if (sxmax > source->width)
        {
        sxmax = source->width;
        }

    for (sy = symin; sy < symax; sy++, cy++)
        {
        spixel = dynim_canvas_pixelptr(source, sxmin, sy);
        pixel = dynim_canvas_pixelptr(canvas, cx, cy);
        memcpy(pixel, spixel, sxmax - sxmin);
        }

    return canvas;
    }

void dynim_canvas_paste(DynImCanvas* canvas, DynImCanvas* source, int x, int y)
    {
    int sx, sy;
    int cxmin, cxmax, cymin, cymax, cx, cy;
    byte* spixel;
    byte* pixel;

    sy = 0;
    cymin = y;
    if (cymin < 0)
        {
        sy += (-cymin);
        cymin = 0;
        }
    cymax = y + source->height;
    if (cymax > canvas->height)
        {
        cymax = canvas->height;
        }

    sx = 0;
    cxmin = x;
    if (cxmin < 0)
        {
        sx += (-cxmin);
        cxmin = 0;
        }
    cxmax = x + source->width;
    if (cxmax > canvas->width)
        {
        cxmax = canvas->width;
        }

    for (cy = cymin; cy < cymax; cy++, sy++)
        {
        spixel = dynim_canvas_pixelptr(source, sx, sy);
        pixel = dynim_canvas_pixelptr(canvas, cxmin, cy);
        memcpy(pixel, spixel, cxmax - cxmin);
        }
    }
