/*
Copyright (C) September 1999, Ka-Ping Yee

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "dynim.h"
#include "table.h"

/* GIF encoding routines =============================================== */

/* --------------------------------------------------------- block types */
void dynim_gif_start(int handle, int width, int height)
    {
    char header[6] = "GIF89a";
    char lsd[7] = "\x00\x00"                    /* width, big-endian */
                  "\x00\x00"                    /* height, big-endian */
                  "\x00"                        /* no gct, no depth */
                  "\x00"                        /* background colour is 0 */
                  "\x00";                       /* no pixel aspect info */

    lsd[0] = width & 0xff;
    lsd[1] = width >> 8;
    lsd[2] = height & 0xff;
    lsd[3] = height >> 8;

    write(handle, header, 6);
    write(handle, lsd, 7);
    }

void dynim_gif_putdata(int handle, byte* data, int size)
    {
    byte chunk;

    while (size > 0)
        {
        chunk = size > 16 ? 16 : size;
        write(handle, &chunk, 1);
        write(handle, data, chunk);

        size -= chunk;
        data += chunk;
        }

    chunk = 0;
    write(handle, &chunk, 1);
    }

void dynim_gif_putappext(int handle, char* appid, char* appauth,
                         byte* data, int size)
    {
    char appext[3] = "\x21\xff"                 /* application ext tag */
                     "\x0b";                    /* size is 11 bytes */

    write(handle, appext, 3);
    write(handle, appid, 8);
    write(handle, appauth, 3);
    dynim_gif_putdata(handle, data, size);
    }

void dynim_gif_putcomment(int handle, char* text)
    {
    char comment[2] = "\x21\xfe";               /* comment tag */

    write(handle, comment, 2);
    dynim_gif_putdata(handle, text, strlen(text));
    }

void dynim_gif_putcanvas(int handle, DynImCanvas* canvas,
                         DynImBuffer* buffer, int x, int y)
    {
    char gcext[8] = "\x21\xf9"                  /* graphic control ext tag */
                    "\x04"                      /* size is 4 bytes */
                    "\x04"                      /* no dispose, use transp */
                    "\x00\x00"                  /* no delay */
                    "\x00"                      /* background colour is 0 */
                    "\x00";                     /* terminator */
    char imgdesc[10] = "\x2c"                   /* image descriptor tag */
                       "\x00\x00"               /* left */
                       "\x00\x00"               /* top */
                       "\x00\x00"               /* width */
                       "\x00\x00"               /* height */
                       "\x00";                  /* flags */
    int c;

    imgdesc[1] = x & 0xff;
    imgdesc[2] = x >> 8;
    imgdesc[3] = y & 0xff;
    imgdesc[4] = y >> 8;
    imgdesc[5] = canvas->width & 0xff;
    imgdesc[6] = canvas->width >> 8;
    imgdesc[7] = canvas->height & 0xff;
    imgdesc[8] = canvas->height >> 8;
    imgdesc[9] = 0x80 + buffer->depth-1;

    write(handle, gcext, 8);
    write(handle, imgdesc, 10);

    for (c = 0; c < (1 << buffer->depth); c++)  /* local colour table */
        {
        if (c < canvas->palette->size)
            write(handle, &(canvas->palette->colours[c]), 3);
        else
            write(handle, "\x00\x00\x00", 3);
        }

    write(handle, &(buffer->depth), 1);
    dynim_gif_putdata(handle, buffer->data, buffer->size);
    }

void dynim_gif_end(int handle)
    {
    char trailer[1] = "\x3b";                   /* GIF trailer */
    write(handle, trailer, 1);
    }

/* --------------------------------------------------------- compression */
Table* table_init(int bits)
    {
    Table* table = table_new(257);
    char data;
    int code;

    for (code = 0; code < (1 << bits); code++)
        {
        data = code;
        table_put(table, &data, 1, code);
        }
    return table;
    }

DynImBuffer* dynim_gif_encode(DynImCanvas* canvas)
    {
    DynImBuffer* buffer;
    Table* table;
    byte* pixel;
    byte* end;
    int lastcode = -1, nextcode = 0, clear, eoi, code;
    int codesize, depth;
    char prefix[1000];
    int prelen = 0;

    for (depth = 0; (1 << depth) < canvas->palette->size; depth++);
    buffer = dynim_buffer_new(256, depth);

    table = table_init(depth);
    nextcode = 1 << depth;
    clear = nextcode++;
    eoi = nextcode++;
    codesize = depth + 1;
    dynim_buffer_putbits(buffer, clear, codesize);

    end = canvas->data + canvas->width * canvas->height;
    for (pixel = canvas->data; pixel < end; pixel++)
        {
        if (prelen)
            {
            prefix[prelen++] = *pixel;
            code = table_get(table, prefix, prelen);
            if (code >= 0) lastcode = code;
            else
                {
                dynim_buffer_putbits(buffer, lastcode, codesize);
                if (nextcode < 4096)
                    {
                    table_put(table, prefix, prelen, nextcode++);
                    if (nextcode > (1 << codesize)) codesize++;
                    }
                prelen = 0;
                }
            }

        if (!prelen) lastcode = prefix[prelen++] = *pixel;
        }

    if (lastcode >= 0) dynim_buffer_putbits(buffer, lastcode, codesize);
    dynim_buffer_putbits(buffer, eoi, codesize);
    table_delete(table);
    return buffer;
    }
